/* skype_dsp_hijacker 0.6 - redirect the open system call to a second dsp device.
 *                        - patch for separate devices
 *                        - workaround a bug introduced in Skype 1.2.0.11
 *
 * Copyright (C) 2004 <snow-x@web.de>
 * Copyright (C) 2005 <jan@nsgroup.net>
 * Copyright (C) 2005-2006 Jan Slupski <jslupski@juljas.net>
 * Copyright (C) 2005 _26oo_ user of Skype Forum
 * Copyright (C) 2006 Romano Giannetti <romano@dea.icai.upcomillas.es>
 *
 * Special thanks to the KDE Project for arts. This program is based on artsdsp.
 * Copyright (C) 1998 Manish Singh <yosh@gimp.org>
 * Copyright (C) 2000 Stefan Westerfeld <stefan@space.twc.de> (aRts port)
 * 
 * Resources:
 *  [http://suedpol.dyndns.org/skype/]
 *  http://195.38.3.142:6502/skype/
 *  http://juljas.net/linux/skype/
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define HIJACKER_VERSION "0.7"

#define _GNU_SOURCE

#include <dlfcn.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
/* #include <ansidecl.h> */

//the device Skype is opening by default (override with environment variable HIJACKDSP)
#define HIJACKDSP "/dev/dsp"
//the mixer Skype is opening by default (override with environment variable HIJACKMIXER)
#define HIJACKMIXER "/dev/mixer"

//debug
// 0 - no messages
// 1 - only very basic messages (recommended)
// 2 - more output on opening/closing sound devices
// 3 - lots of messages on open/close actions
#define DEBUG_HIJACKER 1

//enable workaround bug in Skype 1.2.0.11 where Skype is trying to open dsp
//not closing it before (and failing thus)
#define WANT_DOUBLE_OPEN_FIX 1

//real device that will be used for
//microphone (override with environment variable MICDEV)
#define DEFAULT_MIC "/dev/dsp"
//speakers (override with environment variable SPEAKERDEV)
#define DEFAULT_SPEAKER "/dev/dsp1"
//mixer (override with environment variable MIXERDEV)
#define DEFAULT_MIXER "/dev/mixer1"

/* original C Library functions */
typedef int (*orig_open_ptr)(const char *pathname, int flags, ...);
typedef int (*orig_close_ptr)(int fd);
typedef int (*orig_write_ptr)(int fd, const void *buf, size_t count);
typedef int (*orig_ioctl_ptr)(int d, int request, char *argp);

static orig_open_ptr orig_open;
static orig_close_ptr orig_close;
static orig_write_ptr orig_write;
static orig_ioctl_ptr orig_ioctl;

static int lib_init = 0;
static int micfd = -1, speakerfd = -1, dsp_is_open = 0;
static char *final_mic, *final_speaker, *final_mixer, *dsp, *mixer;
static void skype_dsp_doinit(void);
static void names_init(void);
static void print_config(void);

#define CHECK_INIT() if(!lib_init) { lib_init=1; skype_dsp_doinit(); names_init(); print_config(); }

ssize_t write(int fd, const void *buf, size_t count)
{
	CHECK_INIT();

	//reroute writes to speaker (that can be other device than microphone)
	if(fd == micfd  &&  dsp_is_open){
		fd = speakerfd;
	}

	return orig_write(fd, buf, count);
}


int ioctl (int d, int request, char *argp)
{
	CHECK_INIT();

	if(DEBUG_HIJACKER>2)
		fprintf(stderr,"hijacker: ioctl called with fd %d\n", d);

	//clone ioctl of microphone to speakers as well
	if(d == micfd  &&  dsp_is_open){
		orig_ioctl(micfd, request, argp);
		orig_ioctl(speakerfd, request, argp);
		return 0;
	}
	else
		return orig_ioctl(d, request, argp);
}

int close(int fd)
{
	CHECK_INIT();

	if(DEBUG_HIJACKER>2)
		fprintf(stderr,"hijacker: close called with fd %d\n", fd);

	if (fd == micfd  &&  dsp_is_open){ 
		if(DEBUG_HIJACKER>1){
			fprintf(stderr,"hijacker: close called with mic fd %d\n", fd);
			fprintf(stderr,"hijacker: close speaker %d too\n", speakerfd);
		}
		orig_close(speakerfd);	/* We close the speaker too */	 
		dsp_is_open = 0; 
		speakerfd = -1;
		micfd = -1;
	} 

	return(orig_close(fd));
}

int open (const char *pathname, int flags, ...)
{
	int fd;
	va_list args;
	mode_t mode = 0;

	CHECK_INIT();

	va_start(args,flags);
	if(flags & O_CREAT) {
		if (sizeof(int) >= sizeof(mode_t)) {
			mode = va_arg(args, int);
		} else {
			mode = va_arg(args, mode_t);
		}
	}
	va_end(args);

	/*
	 * If 'open' is trying to open the configured dsp device, it gets a new
	 * pathname of the chosen sound device. The same with the mixer
	 * device.
	 */
	if(strcmp(pathname, dsp) == 0){

		if(DEBUG_HIJACKER>0)
			fprintf(stderr,"hijacker: open DSP %s flags: %d\n",pathname, flags);

#if (WANT_DOUBLE_OPEN_FIX)
		if(micfd!=-1){
			if(DEBUG_HIJACKER>0)
				fprintf(stderr,"hijacker: SKYPE 1.2.0.11 BUG WORKAROUND, open mic %s which wasn't closed. closing it before reopening\n",final_mic);
			close(micfd);
		}  
#endif
		micfd = orig_open(final_mic, O_RDONLY, mode);
		if(DEBUG_HIJACKER>0)
			fprintf(stderr,"hijacker: mic %s opened with fd %d\n", final_mic, micfd);


#if (WANT_DOUBLE_OPEN_FIX)
		if(speakerfd!=-1){
			if(DEBUG_HIJACKER>0)
				fprintf(stderr,"hijacker: SKYPE 1.2.0.11 BUG WORKAROUND, open speaker %s which wasn't closed. closing it before reopening\n",final_speaker);
			close(speakerfd);
		}
#endif
		speakerfd = orig_open(final_speaker, O_WRONLY, mode);
		if(DEBUG_HIJACKER>0)
			fprintf(stderr,"hijacker: speaker %s opened with fd %d\n", final_speaker, speakerfd);

		// If sound device is impossible to open, return -1 so that skype notice that
		if (speakerfd < 0 )
			return speakerfd;

		dsp_is_open = 1;

		//return microphone fd that will be later reference for bot micfd & speakerfd
		return micfd;
	}
	else if(strcmp(pathname, mixer) == 0){

		pathname = final_mixer;
		if(DEBUG_HIJACKER>1)
			fprintf(stderr,"hijacker: mixer %s open\n",pathname);
	}

	/* call the original open command */
	fd=orig_open (pathname, flags, mode);

	if(DEBUG_HIJACKER>1)
		if(pathname == final_mixer)
			fprintf(stderr,"hijacker: mixer open %s returned with fd %d\n", pathname, fd);

	return fd;
}

/* Save the original functions */

static void skype_dsp_doinit()
{
	orig_open = (orig_open_ptr)dlsym(RTLD_NEXT,"open");
	orig_close = (orig_close_ptr)dlsym(RTLD_NEXT,"close");
	orig_write = (orig_write_ptr)dlsym(RTLD_NEXT,"write");
	orig_ioctl = (orig_ioctl_ptr)dlsym(RTLD_NEXT,"ioctl");
}

static void names_init(){

	dsp=getenv("HIJACKDSP");
	if(!dsp) dsp=HIJACKDSP;

	mixer=getenv("HIJACKMIXER");
	if(!mixer) mixer=HIJACKMIXER;

	final_mixer = getenv("MIXERDEV");
	if(!final_mixer) final_mixer = DEFAULT_MIXER;

	final_mic = getenv("MICDEV");
	if(!final_mic) final_mic = DEFAULT_MIC;

	final_speaker = getenv("SPEAKERDEV");
	if(!final_speaker) final_speaker = DEFAULT_SPEAKER;
}

static void print_config(){
#define ONOFF(b) ((b)?"ON":"OFF")
	if(DEBUG_HIJACKER>0){
		fprintf(stderr,"hijacker: skype_dsp_hijacker v%s\n", HIJACKER_VERSION);
		fprintf(stderr,"hijacker: - debug level is %d\n", DEBUG_HIJACKER);
		fprintf(stderr,"hijacker: - double open fix is %s\n", ONOFF(WANT_DOUBLE_OPEN_FIX));
		fprintf(stderr,"hijacker: when Skype opens DSP %s\n",dsp);
		fprintf(stderr,"hijacker: - microphone used: %s\n", final_mic);
		fprintf(stderr,"hijacker: - speakers used:   %s\n", final_speaker);
		fprintf(stderr,"hijacker: when Skype opens mixer %s\n",mixer);
		fprintf(stderr,"hijacker: - mixer used: %s\n", final_mixer);
	}

}
